"""
Unit and regression test for the alphaPredict package.
"""

# Import package, test suite, and other packages as needed
import alphaPredict
import pytest
import sys

def test_alphaPredict_imported():
    """Sample test, will always pass so long as import statement worked."""
    assert "alphaPredict" in sys.modules



"""
Backend of the IDR machine learning predictor. Based partly
on code from Dan Griffith's IDP-Parrot from the Holehouse lab
(specifically the test_unlabeled_data function in train_network.py).
"""

# import packages for predictor
import sys
import os

import numpy as np
import torch
import torch.nn as nn
from torch.utils.data import Dataset, DataLoader


# import modules that predictor depends on
from alphaPredict.backend import encode_sequence
from alphaPredict.backend import brnn_architecture


# set path for location of predictor. Using this in case I want to update the predictor or
# eventually make multiple predictors.
PATH = os.path.dirname(os.path.realpath(__file__))

# Setting predictor equal to location of weighted values.
predictor = "{}/networks/alpha_fold_networkV7_hs100_nL2_200e_all_prot.pt".format(PATH)

##################################################################################################
# hyperparameters used by when metapredict was trained. Manually setting them here for clarity.
##################################################################################################
# This is defined externally so its read in and loaded one time on the initial import
#
'''
network parameters for /networks/alpha_fold_networkV1.pt - 

device = 'cpu'
hidden_size = 10
num_layers = 1
dtype = 'residues'
num_classes = 1
encoding_scheme = 'onehot'
input_size = 20
problem_type = 'regression'

network parameters for: 
/networks/alpha_fold_networkV2_hs20_nL2_200e.pt  

and for: 
/networks/alpha_fold_networkV3_hs20_nL2_1000e.pt- 

and for: 
/networks/alpha_fold_networkV4_hs20_nL2_200e_all_prot.pt

device = 'cpu'
hidden_size = 20
num_layers = 2
dtype = 'residues'
num_classes = 1
encoding_scheme = 'onehot'
input_size = 20
problem_type = 'regression'

and for: 
/networks/alpha_fold_networkV5_hs40_nL5_200e_all_prot.pt

device = 'cpu'
hidden_size = 40
num_layers = 5
dtype = 'residues'
num_classes = 1
encoding_scheme = 'onehot'
input_size = 20
problem_type = 'regression'

for 
/networks/alpha_fold_networkV6_hs40_nL2_200e_all_prot.pt

device = 'cpu'
hidden_size = 40
num_layers = 2
dtype = 'residues'
num_classes = 1
encoding_scheme = 'onehot'
input_size = 20
problem_type = 'regression'


for 
/networks/alpha_fold_networkV7_hs100_nL2_200e_all_prot.pt

device = 'cpu'
hidden_size = 100
num_layers = 2
dtype = 'residues'
num_classes = 1
encoding_scheme = 'onehot'
input_size = 20
problem_type = 'regression'
'''


device = 'cpu'
#hidden_size = 20
# changed to 40 for v5 and v6
# changed to 100 for v7
hidden_size = 100

# 2 layers for v6, v1-4
num_layers = 2

# chaned to 5 for v5
#num_layers = 5

dtype = 'residues'
num_classes = 1
encoding_scheme = 'onehot'
input_size = 20
problem_type = 'regression'

# set location of saved_weights for load_state_dict
saved_weights = predictor

###############################################################################
# Initialize network architecture using previously defined hyperparameters
###############################################################################
brnn_network = brnn_architecture.BRNN_MtM(input_size, hidden_size, num_layers, num_classes, device).to(device)
brnn_network.load_state_dict(torch.load(saved_weights, map_location=torch.device(device)))
###############################################################################


def alpha_predict(sequence,  network=brnn_network, device=device, encoding_scheme=encoding_scheme):
    """
    The actual executing function for predicting the disorder of a sequence using metapredict.
    Returns a list containing predicted disorder values for the input sequence. 

    Arguments
    ---------
    sequence - the amino acid sequence to be predicted

    network - the network used by the predictor. See brnn_architecture BRNN_MtO for more info.

    device - String describing where the network is physically stored on the computer. 
    Should be either 'cpu' or 'cuda' (GPU).

    encoding_scheme - encoding scheme used when metapredict was trained. The encoding scheme was onehot.
    """

    # change sequence to uppercase (just in case)
    sequence = sequence.upper()

    # set seq_vector equal to converted amino acid sequence that is a PyTorch tensor of one-hot vectors
    seq_vector = encode_sequence.one_hot(sequence)
    seq_vector = seq_vector.view(1, len(seq_vector), -1)

    # get output values from the seq_vector based on the network (brnn_network)
    output = network(seq_vector.float()).detach().numpy()[0]

    # make empty list to add in outputs
    output_values = []
    # for the values 'i' in outputs
    for i in output:
        # append each value (which is the predicted disorder value) to output values as a float.
        # round each value to six digits.
        output_values.append(round(float(i), 4))

    # return the prediction
    return output_values



##
## alpha.py
## 
## alpha.py contains all the user-facing function associated with alphaPredict. 
## If a new function is added it should be included
## here and added to the __all__ list
## 

##Handles the primary functions


__all__ =  ['predict', 'graph']
 
import os
import sys

# note - we imort packages below with a leading _ which means they are ignored in the import

#import protfasta to read .fasta files
import protfasta as _protfasta

# import stuff for IDR predictor from backend. Note the 'as _*' hides the imported
# module from the user
#from predictAlpha.backend.parrot_alpha import alpha_predict as _alpha_predict
from alphaPredict.backend.parrot_alpha import alpha_predict as _alpha_predict
from alphaPredict.backend import alpha_tools as _alpha_tools
from alphaPredict.backend.alpha_graph import graph as _graph
# stuff for uniprot from backend
from alphaPredict.alpha_exceptions import AlphaError



def predict(sequence):
    """
    Function to return confidence scores from
    Alpha fold 2 of a single input sequence. Returns the
    predicted values as a float.

    Parameters
    ------------

    sequence : str 
        Input amino acid sequence (as string) to be predicted.

    Returns
    --------
    
    Float
        Returns a float of the confidence score value (predicted)

    """
    # make all residues upper case 
    sequence = sequence.upper()

    # return predicted values of disorder for sequence
    return _alpha_predict(sequence)


def graph(sequence,
          title='Predicted Confidence Score',
          confidence_threshold=50,
          shaded_regions=None,
          shaded_region_color='red',
          confidence_line_color='blue',
          threshold_line_color='black',
          DPI=150,
          output_file=None):
    _graph(sequence, title, confidence_threshold, 
            shaded_regions, shaded_region_color,
            confidence_line_color, threshold_line_color,
            DPI, output_file)
    
    



"""
Backend for graphing predicted disorder values in meta.py.
"""


# code for graphing IDRs.
# Import stuff
import numpy as np
import matplotlib
import matplotlib.pyplot as plt

from alphaPredict import alpha

def graph(sequence,
          title='Predicted Confidence Score',
          confidence_threshold=50,
          shaded_regions=None,
          shaded_region_color='red',
          confidence_line_color='blue',
          threshold_line_color='black',
          DPI=150,
          output_file=None):
    """
    Function for graphing predicted confidence scores
    from alpha fold 2. By default, this function will show a graph.
    However, if saveFig = True, then it will save the figure (by default) to the location
    where the script is (which isn't ideal). However, you can specify outputFile as the
    file path followed by the name of the saved file with the proper extension (.png by default).
    This is the backend for the meta.py graphing functions.

    Parameters
    -----------

    sequence : str 
        Input amino acid sequence (as string) to be predicted.

    title : str
        Sets the title of the generated figure. Default = "Predicted protein disorder"

    confidence_threshold : float
        Sets a threshold which draws a horizontal black line as a visual guide along
        the length of the figure. Must be a value between 0 and 1. Default = 0.3

    shaded_regions : list of lists
        A list of lists, where sub-elements are of length 2 and contain start and end
        values for regions to be shaded. Assumes that sanity checking on positions has
        already been done. Default is None.

    shaded_region_color : str
        String that defines the color of the shaded region. The shaded region is always
        set with an alpha of 0.3 but the color can be any valid matplotlib color name
        or a hex color string (i.e. "#ff0000" is red). Default = 'red'.

    confidence_line_color : str
        String that defines the color of the traced disorder score.  Can
        be any standard matplotlib color name or a hex-value (see above). 
        Default = 'blue'.

    threshold_line_color : str
        String that defines the color of the traced disorder score. Can
        be any standard matplotlib color name or a hex-value (see above). 
        Default = 'black'.

    DPI : int
        Dots-per-inch. Defines the resolution of the generated .png figure. Note that
        if an alternative filetype is pathed the matplotlib backened will automatically
        generate a file of the relevant type (e.g. .pdf, .jpg, or .eps).

    output_file : str
        If provided, the output_file variable defines the location and type of the file
        to be saved. This should be a file location and filename with a valid matplotlib
        extension (such as .png, .jpg, .pdf) and, if provided, this value is passed directly
        to the ``matplotlib.pyplot.savefig()`` function as the ``fname`` parameter. 
        Default = None.


    Returns
    -----------
    None 
        No return type, but will either generate an on-screen plot OR will save a file to disk,
        depending on if output_file is provided (or not).

    """

    # set this such that PDF-generated figures become editable
    matplotlib.rcParams['pdf.fonttype'] = 42
    matplotlib.rcParams['ps.fonttype'] = 42

    # set yValues equal to the predicted disorder from the sequence (normalized)
    yValues = alpha.predict(sequence)
    n_res = len(sequence)

    # if a name is set, the figure will hold that name as the identifier
    fig = plt.figure(num=title, figsize=[8, 3], dpi=DPI, edgecolor='black')
    axes = fig.add_axes([0.15, 0.15, 0.75, 0.75])
    axes.set_title(title)
    axes.set_xlabel("Residue")
    axes.set_ylabel("Predicted Confidence Score")

    # make x values for each residue with predicted disorder
    xValues = np.arange(1, n_res+1)

    # graph the disorder values of each residue at each point along the x-axis
    axes.plot(xValues, yValues, color=confidence_line_color, linewidth='1.6')

    # set x limit as the number of residues
    axes.set_xlim(1, n_res+1)

    # set y limit as 0-1 since the predictor data is normalized from 0 to 1.
    axes.set_ylim(-0.003, 100.003)

    # plot the disorder cutoff threshold h
    axes.plot([0, n_res+2], [confidence_threshold, confidence_threshold], color="black", linewidth="1.25")

    # add dashed lines at 0.2 intervals if cutoff lines not specified
    for i in [20, 40, 60, 80]:
        axes.plot([0, n_res+2], [i, i], color="black", linestyle="dashed", linewidth="0.5")

    # if we want shaded regions
    if shaded_regions is not None:
        for boundaries in shaded_regions:
            start = boundaries[0]
            end = boundaries[1]
            axes.axvspan(start, end, alpha=0.2, color=shaded_region_color)

    if output_file is None:
        plt.show()
    else:
        plt.savefig(fname=output_file, dpi=DPI)
        plt.close()






import re
from alphaPredict.alpha_exceptions import AlphaError



def write_csv(input_dict, output_file):
    """
    Function that writes the scores in an input dictionary out to a standardized CVS file format.

    Parameters
    -----------
    input_dict : dict
        Dictionary where keys are headers/identifiers and values is a list of per-residue
        disorder score

    output_file : str
        Location and filename for the output file. Assumes .csv is provided.

    Returns
    --------
    None
        No return value, but writes a .csv file to disk


    """

    # try and open the file and throw exception if anything goes wrong
    try:
        fh = open(output_file, 'w')
    except Exception:
        raise AlphaError('Unable to write to file destination %s' % (output_file))

    # for each entry
    for idx in input_dict:

        # important otherwise commmas in FASTA headers render the CSV file unreadable!
        no_comma = idx.replace(',', ' ')
        fh.write('%s' % (no_comma))

        # for each score write
        for score in input_dict[idx]:
            fh.write(', %1.3f' % (score))
        fh.write('\n')
